use std::collections::{HashSet, VecDeque};
use std::sync::Arc;

use indexed_db::{Database, Factory};
use offline_web_model::Reference;
use serde::{Deserialize, Serialize};

use crate::ReferenceStore;
use crate::StoreError;

const DB_NAME: &str = "offline_web_store";
const DB_VERSION: u32 = 1;
const REF_ENTRIES_STORE: &str = "ref_entries";
const REF_DEPENDENCIES_STORE: &str = "ref_dependencies";
const CONTENT_STORE: &str = "content_store";

#[derive(Serialize, Deserialize, Clone)]
struct RefEntry {
    id: String,
    content_address: Option<String>,
    name: String,
}

#[derive(Serialize, Deserialize, Clone)]
struct RefDependency {
    parent_id: String,
    dependent_id: String,
}

#[derive(Serialize, Deserialize, Clone)]
struct ContentEntry {
    content_address: String,
    content: Vec<u8>,
}

pub struct IndexedDbReferenceStore {
    db: Database<StoreError>,
}

impl IndexedDbReferenceStore {
    pub async fn new() -> Result<Self, StoreError> {
        let factory = Factory::<StoreError>::get()?;

        let db = factory
            .open(DB_NAME, DB_VERSION, |event| async move {
                let db = event.database();
                
                // Create ref_entries object store
                let ref_store = db
                    .build_object_store(REF_ENTRIES_STORE)
                    .create()?;
                
                // Create name index for get_graph method
                ref_store
                    .build_index("name", "name")
                    .create()?;

                // Create ref_dependencies object store
                db.build_object_store(REF_DEPENDENCIES_STORE)
                    .create()?;

                // Create content_store object store
                db.build_object_store(CONTENT_STORE)
                    .create()?;

                Ok(())
            })
            .await?;

        Ok(Self { db })
    }

    async fn get_reference_without_dependents(&self, id: &str) -> Result<Reference, StoreError> {
        let id = id.to_string();
        let reference = self.db
            .transaction(&[REF_ENTRIES_STORE])
            .run(|transaction| async move {
                let store = transaction.object_store(REF_ENTRIES_STORE)?;
               
                let id_value = serde_wasm_bindgen::to_value(&id)
                        .map_err(|e| {
                            indexed_db::Error::User(StoreError::SerializationError(Box::new(e)))
                        })?;
                let value = store.get(&id_value).await?;

                if let Some(js_value) = value {
                    let ref_entry: RefEntry = serde_wasm_bindgen::from_value(js_value)
                        .map_err(|e| {
                            indexed_db::Error::User(StoreError::SerializationError(Box::new(e)))
                        })?;

                    Ok(Reference {
                        id: ref_entry.id,
                        content_address: ref_entry.content_address,
                        name: ref_entry.name,
                        dependents: Vec::new(),
                    })
                } else {
                    Err(indexed_db::Error::User(StoreError::NoSuchReference))
                }
            })
            .await?;
        Ok(reference)
    }

    async fn get_direct_children(&self, parent_id: &str) -> Result<Vec<String>, StoreError> {
        let parent_id = parent_id.to_string();
        self.db
            .transaction(&[REF_DEPENDENCIES_STORE])
            .run(|transaction| async move {
                let store = transaction.object_store(REF_DEPENDENCIES_STORE)?;
                let mut children = Vec::new();
                
                let mut cursor = store.cursor().open().await?;
                
                while let Some(value) = cursor.value() {
                    if let Ok(dep_entry) = serde_wasm_bindgen::from_value::<RefDependency>(value) {
                        if dep_entry.parent_id == parent_id {
                            children.push(dep_entry.dependent_id);
                        }
                    }
                    cursor.advance(1).await?;
                }
                
                Ok(children)
            })
            .await.map_err(StoreError::from)
    }

    async fn get_dependents(&self, parent_id: &str) -> Result<Vec<Arc<Reference>>, StoreError> {
        let mut all_refs = std::collections::HashMap::new();
        let mut dependency_map = std::collections::HashMap::<String, Vec<String>>::new();
        let mut to_process = std::collections::VecDeque::new();
        let mut processed = std::collections::HashSet::new();
        
        let direct_children = self.get_direct_children(parent_id).await?;
        for child_id in direct_children.clone() {
            to_process.push_back(child_id);
        }
        
        while let Some(ref_id) = to_process.pop_front() {
            if processed.contains(&ref_id) {
                continue;
            }
            processed.insert(ref_id.clone());
            
            if let Ok(reference) = self.get_reference_without_dependents(&ref_id).await {
                all_refs.insert(ref_id.clone(), reference);
                
                if let Ok(children) = self.get_direct_children(&ref_id).await {
                    dependency_map.insert(ref_id.clone(), children.clone());
                    for child_id in children {
                        to_process.push_back(child_id);
                    }
                } else {
                    dependency_map.insert(ref_id.clone(), Vec::new());
                }
            }
        }
        
        let mut result = Vec::new();
        for child_id in direct_children {
            if let Some(reference) = self.build_full_reference(&child_id, &all_refs, &dependency_map) {
                result.push(Arc::new(reference));
            }
        }
        
        Ok(result)
    }
    
    fn build_full_reference(
        &self,
        target_id: &str,
        all_refs: &std::collections::HashMap<String, Reference>,
        dependency_map: &std::collections::HashMap<String, Vec<String>>
    ) -> Option<Reference> {
        let mut built_refs = std::collections::HashMap::<String, Reference>::new();
        let mut to_build = std::collections::VecDeque::new();
        let mut processed = std::collections::HashSet::new();
        
        to_build.push_back(target_id.to_string());
        
        while let Some(ref_id) = to_build.pop_back() {
            if processed.contains(&ref_id) {
                continue;
            }
            
            let base_ref = all_refs.get(&ref_id)?;
            let children = dependency_map.get(&ref_id).cloned().unwrap_or_default();
            
            let all_children_built = children.iter().all(|child_id| built_refs.contains_key(child_id));
            
            if all_children_built {
                let mut dependents = Vec::new();
                for child_id in &children {
                    if let Some(child_ref) = built_refs.get(child_id) {
                        dependents.push(Arc::new(child_ref.clone()));
                    }
                }
                
                let full_ref = Reference {
                    id: base_ref.id.clone(),
                    content_address: base_ref.content_address.clone(),
                    name: base_ref.name.clone(),
                    dependents,
                };
                
                built_refs.insert(ref_id.clone(), full_ref);
                processed.insert(ref_id);
            } else {
                to_build.push_front(ref_id);
                for child_id in children {
                    if !processed.contains(&child_id) && !built_refs.contains_key(&child_id) {
                        to_build.push_back(child_id);
                    }
                }
            }
        }
        
        built_refs.remove(target_id)
    }

    async fn find_reference_by_name(&self, name: &str) -> Result<Option<RefEntry>, StoreError> {
        let name = name.to_string();
        
        self.db
            .transaction(&[REF_ENTRIES_STORE])
            .run(|transaction| async move {
                let store = transaction.object_store(REF_ENTRIES_STORE)?;
                let mut cursor = store.cursor().open().await?;
                
                while let Some(value) = cursor.value() {
                    if let Ok(ref_entry) = serde_wasm_bindgen::from_value::<RefEntry>(value) {
                        if ref_entry.name == name {
                            return Ok(Some(ref_entry));
                        }
                    }
                    cursor.advance(1).await?;
                }
                
                Ok(None)
            })
            .await.map_err(StoreError::from)
    }
}

impl ReferenceStore for IndexedDbReferenceStore {
    async fn get_reference(&self, id: &str) -> Result<Reference, StoreError> {
        let mut reference = self.get_reference_without_dependents(id).await?;
        reference.dependents = self.get_dependents(id).await?;
        Ok(reference)
    }

    async fn get_content_for_reference(&self, reference: Reference) -> Result<String, StoreError> {
        if let Some(content_address) = reference.content_address {
            self.db
                .transaction(&[CONTENT_STORE])
                .run(|transaction| async move {
                    let store = transaction.object_store(CONTENT_STORE)?;
                    
                    let content_key = serde_wasm_bindgen::to_value(&content_address)
                        .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                    let value = store.get(&content_key).await?;

                    if let Some(js_value) = value {
                        let content_entry: ContentEntry = serde_wasm_bindgen::from_value(js_value)
                            .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;

                        String::from_utf8(content_entry.content)
                            .map_err(|e| indexed_db::Error::User(StoreError::StorageError(Box::new(e))))
                    } else {
                        Err(indexed_db::Error::User(StoreError::NoSuchContentAddress))
                    }
                })
                .await.map_err(StoreError::from)
        } else {
            Err(StoreError::NoSuchContentAddress)
        }
    }

    async fn get_graph(&self, root_name: &str) -> Result<Vec<Reference>, StoreError> {
        let mut visited = HashSet::new();
        let mut result = Vec::new();
        let mut queue = VecDeque::new();

        queue.push_back(root_name.to_string());

        while let Some(current_name) = queue.pop_front() {
            if visited.contains(&current_name) {
                continue;
            }

            visited.insert(current_name.clone());

            if let Ok(reference_opt) = self.find_reference_by_name(&current_name).await {
                if let Some(ref_entry) = reference_opt {
                    let dependents = self.get_dependents(&ref_entry.id).await?;

                    let reference = Reference {
                        id: ref_entry.id,
                        content_address: ref_entry.content_address,
                        name: ref_entry.name,
                        dependents: dependents.clone(),
                    };

                    result.push(reference);

                    for dependent in dependents {
                        if !visited.contains(&dependent.name) {
                            queue.push_back(dependent.name.clone());
                        }
                    }
                }
            }
        }

        Ok(result)
    }
    
    async fn store_reference(&self, reference: &Reference) -> Result<(), StoreError> {
        let reference = reference.clone();
        
        self.db
            .transaction(&[REF_ENTRIES_STORE, REF_DEPENDENCIES_STORE])
            .rw()
            .run(|transaction| async move {
                let ref_store = transaction.object_store(REF_ENTRIES_STORE)?;
                let dep_store = transaction.object_store(REF_DEPENDENCIES_STORE)?;

                // Store the reference entry
                let ref_entry = RefEntry {
                    id: reference.id.clone(),
                    content_address: reference.content_address.clone(),
                    name: reference.name.clone(),
                };

                let ref_value = serde_wasm_bindgen::to_value(&ref_entry)
                    .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                let ref_key = serde_wasm_bindgen::to_value(&reference.id)
                    .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                ref_store.add_kv(&ref_key, &ref_value).await?;

                // Store new dependencies
                for dependent in &reference.dependents {
                    let dep_entry = RefDependency {
                        parent_id: reference.id.clone(),
                        dependent_id: dependent.id.clone(),
                    };

                    let dep_value = serde_wasm_bindgen::to_value(&dep_entry)
                        .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                    let dep_key = serde_wasm_bindgen::to_value(&format!("{}:{}", reference.id, dependent.id))
                        .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                    
                    dep_store.add_kv(&dep_key, &dep_value).await?;
                }

                Ok(())
            })
            .await.map_err(StoreError::from)
    }
    
    async fn store_content(&self, content_address: &str, content: &[u8]) -> Result<(), StoreError> {
        let content_address = content_address.to_string();
        let content = content.to_vec();
        
        self.db
            .transaction(&[CONTENT_STORE])
            .rw()
            .run(|transaction| async move {
                let store = transaction.object_store(CONTENT_STORE)?;

                let content_entry = ContentEntry {
                    content_address: content_address.clone(),
                    content,
                };

                let content_value = serde_wasm_bindgen::to_value(&content_entry)
                    .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                let content_key = serde_wasm_bindgen::to_value(&content_entry.content_address)
                    .map_err(|e| indexed_db::Error::User(StoreError::SerializationError(Box::new(e))))?;
                store.add_kv(&content_key, &content_value).await?;

                Ok(())
            })
            .await.map_err(StoreError::from)
    }
}
