use thiserror::Error;
use offline_web_model::Reference;

#[derive(Error, Debug)]
pub enum StoreError {
    #[error("No such reference")]
    NoSuchReference,
    #[error("Invalid reference")]
    InvalidReference,
    #[error("No such content address")]
    NoSuchContentAddress,
    #[error("Serialization Error: {0:?}")]
    SerializationError(Box<dyn std::error::Error>),
    #[error("Unknown Storage Error: {0:?}")]
    StorageError(Box<dyn std::error::Error>),
}

#[cfg(feature="wasm")]
impl From<indexed_db::Error<StoreError>> for StoreError {
    fn from(value: indexed_db::Error<StoreError>) -> Self {
        match value {
            indexed_db::Error::User(store_error) => store_error,
            other => StoreError::StorageError(Box::new(other))
        }
    }
}

#[cfg(feature="wasm")]
impl From<serde_wasm_bindgen::Error> for StoreError {
    fn from(value: serde_wasm_bindgen::Error) -> Self {
        StoreError::SerializationError(Box::new(value))
    }
}

impl From<serde_json::Error> for StoreError {
    fn from(value: serde_json::Error) -> Self {
        StoreError::SerializationError(Box::new(value))
    }
}

#[allow(async_fn_in_trait)]
pub trait ReferenceStore {
    async fn get_reference(&self, id: &str) -> Result<Reference, StoreError>;
    
    async fn get_content_for_reference(&self, reference: Reference) -> Result<String, StoreError>;

    async fn get_graph(&self, root_name: &str) -> Result<Vec<Reference>, StoreError>;
    
    async fn store_reference(&self, reference: &Reference) -> Result<(), StoreError>;
    
    async fn store_content(&self, content_address: &str, content: &[u8]) -> Result<(), StoreError>;
}

#[cfg(feature="native")]
mod sqlite;

#[cfg(feature="native")]
pub use sqlite::SqliteReferenceStore;

#[cfg(feature="wasm")]
mod indexeddb;

#[cfg(feature="wasm")]
pub use indexeddb::IndexedDbReferenceStore;

#[cfg(test)]
mod integration_tests;
